/*
 * Usage:  mkfs [-c] [-nXX] [-iXX] [device size-in-blocks]
 *         mkfs [-l filename ] [device size-in-blocks]
 *
 *	-n for namelength (currently the kernel only uses 14 or 30)
 *	-i for number of inodes
 *        
 *       Typical usage: 
 *          mkfs.viva /dev/hda3 - will find the partition size automatically
 *                                and use it, and will allow 14 character
 *                                long file names. 
 *        or mkfs.viva -n 30 /dev/hda3 - will use allow 30 character file names.
 *
 *    Author: Shankar Pasupathy, July 1996.
 *            Modified by Vamsi Krishna, July 1997.
 *    Lot of code borrowed from mkfs.minix.c (c) Linux Torvalds.
 *    This program may be redistributed as per the GNU copyleft.
 */

/* IMPORTANT:
This version of mkfs uses BLOCKS starting at 0, not at bit 1 of the bitmap
as MINIX does. This means that a lot of code was changed. The Badblocks 
code needs to be changed as well, to use bit 0, but I don't have the time 
to do it right now.
*/

#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/stat.h>
#include <mntent.h>
#include <math.h>
#include <sys/ioctl.h>

#include "fs.h"
#include "../../viva_fs.h"

#ifndef __GNUC__
#error "needs gcc for the bitop-__asm__'s"
#endif

#ifndef __linux__
#define volatile
#endif

#define VIVA_ROOT_INO 1
#define VIVA_BAD_INO 2

#define TEST_BUFFER_BLOCKS 16
#define MAX_GOOD_BLOCKS 512

#define UPPER(size,n) ((size+((n)-1))/(n))
#define INODE_SIZE (sizeof(struct viva_inode))
#define INODE_BLOCKS UPPER(INODES,VIVA_INODES_PER_BLOCK)
#define INODE_BUFFER_SIZE (INODE_BLOCKS * VIVA_BLOCK_SIZE)

#define BITS_PER_BLOCK (VIVA_BLOCK_SIZE<<3)

static char * program_name = "mkfs";
static char * device_name = NULL;
static int DEV = -1;
static long BLOCKS = 0;
static long frag_blocks = 0;
static int check = 0;
static int badblocks = 0;
static int namelen = 30;	/* default (changed to 30, per Linus's
				   suggestion, Sun Nov 21 08:05:07 1993) */
static int dirsize = 16;
static int magic = VIVA_SUPER_MAGIC;

static char root_block[VIVA_BLOCK_SIZE] = "\0";

static char * inode_buffer = NULL;
#define Inode (((struct viva_inode *) inode_buffer)-1)
static char super_block_buffer[VIVA_BLOCK_SIZE];
static char boot_block_buffer[512];
#define Super (*(struct viva_super_block *)super_block_buffer)
#define INODES ((unsigned long)Super.s_ninodes)
#define ZONES ((unsigned long)Super.s_nzones)
#define FRAG_BLOCKS ((unsigned long)Super.s_max_frag_blocks)
#define IMAPS ((unsigned long)Super.s_imap_blocks)
#define ZMAPS ((unsigned long)Super.s_zmap_blocks)
#define FS_MAPS ((unsigned long)Super.s_fsmap_blocks)
#define FA_MAPS ((unsigned long)Super.s_famap_blocks)
#define FIRSTZONE ((unsigned long)Super.s_firstdatazone)
#define ZONESIZE ((unsigned long)Super.s_log_zone_size)
#define MAGIC (Super.s_magic)
#define NORM_FIRSTZONE (2+IMAPS+ZMAPS+FS_MAPS+FA_MAPS+INODE_BLOCKS)

/* Vamsi: Not sure about the Size of these Arrays */
static char inode_map[VIVA_BLOCK_SIZE * VIVA_I_MAP_SLOTS];
static char zone_map[VIVA_BLOCK_SIZE * VIVA_Z_MAP_SLOTS];
static char frag_map[VIVA_BLOCK_SIZE * VIVA_FS_MAP_SLOTS];
/* Vamsi: Should be VIVA_FRAG_PER_BLOCK - 1, i guess */
static char frag_alloc_map[VIVA_BLOCK_SIZE * VIVA_FA_MAP_SLOTS];

static unsigned long good_blocks_table[MAX_GOOD_BLOCKS];
static int used_good_blocks = 0;
static unsigned long req_nr_inodes = 0;

#define bitop(name,op) \
static inline int name(char * addr,unsigned int nr) \
{ \
int __res; \
__asm__ __volatile__("bt" op " %1,%2; adcl $0,%0" \
:"=g" (__res) \
:"r" (nr),"m" (*(addr)),"0" (0)); \
return __res; \
}

bitop(bit,"")
bitop(setbit,"s")
bitop(clrbit,"r")

#define inode_in_use(x) (bit(inode_map,(x)))
#define zone_in_use(x) (bit(zone_map,(x)-FIRSTZONE))
#define frag_block_in_use(x) (bit(frag_map,(x)-FIRSTZONE))

#define mark_inode(x) (setbit(inode_map,(x)))
#define unmark_inode(x) (clrbit(inode_map,(x)))

#define mark_zone(x) (setbit(zone_map,(x)-FIRSTZONE))
#define unmark_zone(x) (clrbit(zone_map,(x)-FIRSTZONE))

#define mark_frag_block(x) (setbit(frag_map,(x)-FIRSTZONE))
#define unmark_frag_block(x) (clrbit(frag_map,(x)-FIRSTZONE))

#define mark_frag_alloc(x) (setbit(frag_alloc_map, ((x)- (FIRSTZONE*VIVA_FRAG_PER_BLOCK))))
#define unmark_frag_alloc(x) (clrbit(frag_alloc_map, ((x)- (FIRSTZONE*VIVA_FRAG_PER_BLOCK))))

/*
 * Volatile to let gcc know that this doesn't return. When trying
 * to compile this under viva, volatile gives a warning, as
 * exit() isn't defined as volatile under viva.
 */
volatile void fatal_error(const char * fmt_string,int status)
{
	fprintf(stderr,fmt_string,program_name,device_name);
	exit(status);
}

#define usage() fatal_error("Usage: %s [-c | -l filename] [-nXX] [-iXX] /dev/name blocks\n",16)
#define die(str) fatal_error("%s: " str "\n",8)

/*
 * Check to make certain that our new filesystem won't be created on
 * an already mounted partition.  Code adapted from mke2fs, Copyright
 * (C) 1994 Theodore Ts'o.  Also licensed under GPL.
 */
static void check_mount(void)
{
	FILE * f;
	struct mntent * mnt;

	if ((f = setmntent (MOUNTED, "r")) == NULL)
		return;
	while ((mnt = getmntent (f)) != NULL)
		if (strcmp (device_name, mnt->mnt_fsname) == 0)
			break;
	endmntent (f);
	if (!mnt)
		return;

	die("%s is mounted; will not make a filesystem here!");
}

void write_tables(void)
{
	/* Mark the super block valid. */
	Super.s_state = VIVA_VALID_FS;
	Super.s_state &= ~VIVA_ERROR_FS;

	if (lseek(DEV, 0, SEEK_SET))
		die("seek to boot block failed in write_tables");
	if (512 != write(DEV, boot_block_buffer, 512))
		die("unable to clear boot sector");
	if (VIVA_BLOCK_SIZE != lseek(DEV, VIVA_BLOCK_SIZE, SEEK_SET))
		die("seek failed in write_tables");
	if (VIVA_BLOCK_SIZE != write(DEV, super_block_buffer, VIVA_BLOCK_SIZE))
		die("unable to write super-block");
	if (IMAPS*VIVA_BLOCK_SIZE != write(DEV,inode_map,IMAPS*VIVA_BLOCK_SIZE))
		die("unable to write inode map");
	if (ZMAPS*VIVA_BLOCK_SIZE != write(DEV,zone_map,ZMAPS*VIVA_BLOCK_SIZE))
		die("unable to write zone map");
	if (FS_MAPS*VIVA_BLOCK_SIZE != write(DEV,frag_map,FS_MAPS*VIVA_BLOCK_SIZE))
		die("unable to write fragment map");
	if (FA_MAPS*VIVA_BLOCK_SIZE != 
							write(DEV,frag_alloc_map,FA_MAPS*VIVA_BLOCK_SIZE))
		die("unable to write fragment allocation map");
	if (INODE_BUFFER_SIZE != write(DEV,inode_buffer,INODE_BUFFER_SIZE))
		die("unable to write inodes");
	
}

void write_block(int blk, char * buffer)
{
	if (blk*VIVA_BLOCK_SIZE != lseek(DEV, blk*VIVA_BLOCK_SIZE, SEEK_SET))
		die("seek failed in write_block");
	if (VIVA_BLOCK_SIZE != write(DEV, buffer, VIVA_BLOCK_SIZE))
		die("write failed in write_block");
}

/*
unsigned long partition_size(char *dev_name)
{
	FILE *fp;
	unsigned long psize;
	char s[100];	
	sprintf(s,"fdisk -s > /tmp/viva.tmp %s",dev_name);
	system(s);
	fp=fopen("/tmp/viva.tmp","r");
	fscanf(fp,"%lu",&psize);
	fclose(fp);
	system("rm /tmp/viva.tmp");
	return(psize);
}
*/
unsigned long partition_size(const char *dev_name)
{
	int fd;
	unsigned long psize;
	fd = open(dev_name, O_RDWR);
	if (fd < 0)
		die("unable to open %s while checking partition size");
	if (ioctl(fd, BLKGETSIZE, &psize) < 0)
		die("ioctl failed while checking partition size");
	close(fd);
	psize = psize/(VIVA_BLOCK_SIZE/1024);
	return(psize);
}

int get_free_block(void)
{
	int blk;

	if (used_good_blocks+1 >= MAX_GOOD_BLOCKS)
		die("too many bad blocks");
	if (used_good_blocks)
		blk = good_blocks_table[used_good_blocks-1];
	else
		blk = FIRSTZONE;
	while (blk < ZONES && zone_in_use(blk))
		blk++;
	if (blk >= ZONES)
		die("not enough good blocks");
	good_blocks_table[used_good_blocks] = blk;
	used_good_blocks++;
	return blk;
}

void mark_good_blocks(void)
{
	int blk;

	for (blk=0 ; blk < used_good_blocks ; blk++)
		mark_zone(good_blocks_table[blk]);
}

inline int next(int zone)
{
	if (!zone)
		zone = FIRSTZONE-1;
	while (++zone < ZONES)
		if (zone_in_use(zone))
			return zone;
	return 0;
}

#if 0
void make_bad_inode(void)
{
	struct viva_inode * inode = &Inode[VIVA_BAD_INO];
	int i,j,zone;
	int ind=0,dind=0;
	unsigned short ind_block[VIVA_BLOCK_SIZE>>1];
	unsigned short dind_block[VIVA_BLOCK_SIZE>>1];

#define NEXT_BAD (zone = next(zone))
	if (!badblocks)
		return;
	mark_inode(VIVA_BAD_INO);
	inode->i_nlinks = 1;
	inode->i_time = time(NULL);
	inode->i_mode = S_IFREG + 0000;
	inode->i_size = badblocks*VIVA_BLOCK_SIZE;
	zone = next(0);
	for (i=0 ; i<7 ; i++) {
		inode->i_zone[i] = zone;
		if (!NEXT_BAD)
			goto end_bad;
	}
	inode->i_zone[7] = ind = get_free_block();
	memset(ind_block,0,VIVA_BLOCK_SIZE);
	for (i=0 ; i<512 ; i++) {
		ind_block[i] = zone;
		if (!NEXT_BAD)
			goto end_bad;
	}
	inode->i_zone[8] = dind = get_free_block();
	memset(dind_block,0,VIVA_BLOCK_SIZE);
	for (i=0 ; i<512 ; i++) {
		write_block(ind,(char *) ind_block);
		dind_block[i] = ind = get_free_block();
		memset(ind_block,0,VIVA_BLOCK_SIZE);
		for (j=0 ; j<512 ; j++) {
			ind_block[j] = zone;
			if (!NEXT_BAD)
				goto end_bad;
		}
	}
	die("too many bad blocks");
end_bad:
	if (ind)
		write_block(ind, (char *) ind_block);
	if (dind)
		write_block(dind, (char *) dind_block);
}
#endif

void make_root_inode(void)
{
	struct viva_inode * inode = &Inode[VIVA_ROOT_INO];
	struct viva_smap_entry *entry;

	mark_inode(VIVA_ROOT_INO);
	entry = (struct viva_smap_entry *) (inode->i_smap);
	/* Clear the sparse bitmap */
	/* Vamsi: sizeof(i_smap + i_indirect_block) = 219 */
	memset(inode->i_smap,0,VIVA_DIRECT_SMAP_SIZE+4); 
	inode->i_lastblock = entry->start_addr = get_free_block();
	inode->i_blocksInUse = 1;
	inode->i_nlinks = 2;
	inode->i_time = time(NULL);
	if (badblocks)
		inode->i_size = 3*dirsize;
	else {
		root_block[2*dirsize] = '\0';
		root_block[2*dirsize+1] = '\0';
		inode->i_size = 2*dirsize;
	}
	inode->i_mode = S_IFDIR + 0755;
	inode->i_uid = 304;
	inode->i_gid = 304;
	write_block(inode->i_lastblock,root_block);
}

void setup_tables(void)
{
	unsigned long i;
	unsigned long inode_bitmap_size, zone_bitmap_size;
	unsigned short inode_nsegments, zone_nsegments;
	unsigned short inode_seg_size, zone_seg_size;
	unsigned short inode_lastseg_size, zone_lastseg_size;
	double ceil(double);
	double sqrt(double);


	memset(inode_map,0xff,sizeof(inode_map));
	memset(zone_map,0xff,sizeof(zone_map));
	memset(super_block_buffer,0, VIVA_BLOCK_SIZE);
	memset(boot_block_buffer,0,512);
	MAGIC = magic;
	ZONESIZE = 0;
	ZONES = BLOCKS;
	FRAG_BLOCKS = frag_blocks;
	/* CHANGE HERE - ZONES AND BLOCKS ARE THE SAME with this mkviva */
	/* some magic nrs: 1 inode / 8 blocks */
	if ( req_nr_inodes == 0 ) 
		INODES = BLOCKS/8;
		/* CHANGE HERE */
		/* change this to some other reasonable value */
	else
		INODES = req_nr_inodes;
	/* I don't want some off-by-one errors, so this hack... */
	if ((INODES & 8191) > 8188)
		INODES -= 5;
	if ((INODES & 8191) < 10)
		INODES -= 20;
	IMAPS = UPPER(INODES,BITS_PER_BLOCK);
	ZMAPS = 0;
	while (ZMAPS != UPPER(BLOCKS - NORM_FIRSTZONE,BITS_PER_BLOCK))
		ZMAPS = UPPER(BLOCKS - NORM_FIRSTZONE,BITS_PER_BLOCK);
	FS_MAPS = ZMAPS;
	FA_MAPS = UPPER(FRAG_BLOCKS * VIVA_FRAG_PER_BLOCK,BITS_PER_BLOCK);
	FIRSTZONE = NORM_FIRSTZONE;
	for (i = FIRSTZONE ; i<ZONES ; i++)
		unmark_zone(i); /* 0 for free, 1 for busy */
	for (i = VIVA_ROOT_INO ; i<INODES ; i++)
		unmark_inode(i);
	for (i = FIRSTZONE; i<ZONES; i++)
		unmark_frag_block(i);
	for (i = (FIRSTZONE*VIVA_FRAG_PER_BLOCK); 
					i<=((ZONES-1)*VIVA_FRAG_PER_BLOCK); i++)
		unmark_frag_alloc(i);
	printf("\nSize of Inode bitmap in blocks = %d",IMAPS);
	printf("\nSize of Block bitmap in blocks = %d",ZMAPS);
	printf("\nSize of Fragment bitmap in blocks = %d",FS_MAPS);
	printf("\nSize of Fragment_alloc bitmap in blocks = %d\n",FA_MAPS);
	/* size of inode_bitmap in bytes */
	inode_bitmap_size = (unsigned long)ceil((double)INODES/(double)8);

	/* size of zone bitmap in bytes */
	zone_bitmap_size =(unsigned long)ceil((double)(BLOCKS-FIRSTZONE)/(double)8);

#if 0
	/* Now recompute the number of zones (blocks). Insidious problem
	because segments count atleast a byte of the zone bitmap */
	/* Magic calculation - got a headache trying to figure this one out */
	ZONES = BLOCKS = zone_bitmap_size * 8 + FIRSTZONE;
#endif
	printf("Size of inode bitmap in bytes = %d\n",inode_bitmap_size);
	printf("Size of block bitmap in bytes = %d\n",zone_bitmap_size);

	/* calculate number of segments required for inodes and zones */
	inode_nsegments = (unsigned long)(ceil(sqrt((double)inode_bitmap_size)));
	zone_nsegments = (unsigned long)(ceil(sqrt((double)zone_bitmap_size)));
	if (inode_nsegments > VIVA_MAX_INODE_SEGMENTS) {
		die("Too many inode segments");
	}
	if (zone_nsegments > VIVA_MAX_ZONE_SEGMENTS) {
		die("Too many zone segments");
	}

	Super.s_inode_nsegments = (unsigned short)inode_nsegments;
	Super.s_zone_nsegments = (unsigned short)zone_nsegments;

	inode_seg_size = (unsigned short)ceil((double)inode_bitmap_size/(double)inode_nsegments);
	zone_seg_size = (unsigned short)ceil((double)zone_bitmap_size/(double)zone_nsegments);

	Super.s_inode_seg_size = (unsigned short)inode_seg_size;
	Super.s_zone_seg_size = (unsigned short)zone_seg_size;

	inode_lastseg_size = inode_bitmap_size - (inode_nsegments-1) * inode_seg_size;
	zone_lastseg_size = zone_bitmap_size - (zone_nsegments-1) * zone_seg_size;

	Super.s_inode_lastseg_size = (unsigned short)inode_lastseg_size;
	Super.s_zone_lastseg_size = (unsigned short)zone_lastseg_size;

	printf("Number of segments for inodes = %d\n",inode_nsegments);
	printf("Size of each segment = %d\n", inode_seg_size);
	printf("Size of last segment = %d\n", inode_lastseg_size);

	printf("Number of segments for blocks = %d\n",zone_nsegments);
	printf("Size of each segment = %d\n", zone_seg_size);
	printf("Size of last segment = %d\n", zone_lastseg_size);

/* Vamsi: Removed all the unwanted information stored in the Super Block */

	inode_buffer = malloc(INODE_BUFFER_SIZE);
	if (!inode_buffer)
		die("unable to allocate buffer for inodes");
	memset(inode_buffer,0,INODE_BUFFER_SIZE);
	printf("Created a VIVA file system with:\n");
	printf("%d inodes\n",INODES);
	printf("%d blocks\n",ZONES);
	printf("%d frag_blocks\n",FRAG_BLOCKS);
	printf("First data block is %d (%d)\n",FIRSTZONE,NORM_FIRSTZONE);
	printf("Size of each block in bytes = %d\n",VIVA_BLOCK_SIZE<<ZONESIZE);
}

/*
 * Perform a test of a block; return the number of
 * blocks readable/writeable.
 */
long do_check(char * buffer, int try, unsigned int current_block) 
{
	long got;
	
	/* Seek to the correct loc. */
	if (lseek(DEV, current_block * VIVA_BLOCK_SIZE, SEEK_SET) !=
                       current_block * VIVA_BLOCK_SIZE ) {
                 die("seek failed during testing of blocks");
	}


	/* Try the read */
	got = read(DEV, buffer, try * VIVA_BLOCK_SIZE);
	if (got < 0) got = 0;	
	if (got & (VIVA_BLOCK_SIZE - 1 )) {
		printf("Weird values in do_check: probably bugs\n");
	}
	got /= VIVA_BLOCK_SIZE;
	return got;
}

static unsigned int currently_testing = 0;

void alarm_intr(int alnum)
{
	if (currently_testing >= ZONES)
		return;
	signal(SIGALRM,alarm_intr);
	alarm(5);
	if (!currently_testing)
		return;
	printf("%d ...", currently_testing);
	fflush(stdout);
}

void check_blocks(void)
{
	int try,got;
	static char buffer[VIVA_BLOCK_SIZE * TEST_BUFFER_BLOCKS];

	currently_testing=0;
	signal(SIGALRM,alarm_intr);
	alarm(5);
	while (currently_testing < ZONES) {
		if (lseek(DEV,currently_testing*VIVA_BLOCK_SIZE,SEEK_SET) !=
		currently_testing*VIVA_BLOCK_SIZE)
			die("seek failed in check_blocks");
		try = TEST_BUFFER_BLOCKS;
		if (currently_testing + try > ZONES)
			try = ZONES-currently_testing;
		got = do_check(buffer, try, currently_testing);
		currently_testing += got;
		if (got == try)
			continue;
		if (currently_testing < FIRSTZONE)
			die("bad blocks before data-area: cannot make fs");
		mark_zone(currently_testing);
		badblocks++;
		currently_testing++;
	}
	if (badblocks)
		printf("%d bad block%s\n",badblocks,(badblocks>1)?"s":"");
}

void get_list_blocks(filename)
char *filename;
{
	FILE *listfile;
	unsigned long blockno;

	listfile=fopen(filename,"r");
	if(listfile == (FILE *)NULL) {
		die("can't open file of bad blocks");
	}
	while(!feof(listfile)) {
		fscanf(listfile,"%d\n", &blockno);
		mark_zone(blockno);
		badblocks++;
	}
	if(badblocks) {
		printf("%d bad block%s\n", badblocks, (badblocks>1)?"s":"");
	}
}

int main(int argc, char ** argv)
{
	int i;
	char * tmp;
	struct stat statbuf;
	char * listfile = NULL;
	unsigned long psize;
	int frag_percent = 20; /* Default Percentage of Fragment Blocks allowed */

	if (argc && *argv)
		program_name = *argv;

	printf("\n  INODE SIZE = %d", sizeof(struct viva_inode));
	printf("\n  VIVA_INODES_PER_BLOCK = %d", VIVA_INODES_PER_BLOCK);
	printf("\n  VIVA BLOCK SIZE = %d", VIVA_BLOCK_SIZE);

	if (INODE_SIZE * VIVA_INODES_PER_BLOCK != VIVA_BLOCK_SIZE)
		die("bad inode size");
	while (argc-- > 1) {
		argv++;
		if (argv[0][0] != '-')
			if (device_name) {
				BLOCKS = strtol(argv[0],&tmp,0);
				if (*tmp) {
					printf("strtol error: number of"
					       " blocks not specified correctly");
					usage();
				}
			} else
				device_name = argv[0];
		else { 
			if(argv[0][1] == 'l') {
				listfile = argv[1];
				argv++;
				if (!(argc--))
					usage();
			} else {
				if(argv[0][1] == 'i') {
					req_nr_inodes
					      = (unsigned long)atol(argv[1]);
					argv++;
					if (!(argc--))
						usage();
				} else while (*(++argv[0])) {
					switch (argv[0][0]) {
						case 'c': check=1; break;
						case 'f': frag_percent = strtoul(argv[0]+1,&tmp,0);
								  /* Max Percent of fragments allowed */
								  argv[0][1] = '\0';
								  break;
						case 'n':
							i = strtoul(argv[0]+1,&tmp,0);
							if (*tmp)
								usage();

							argv[0][1] = '\0';
							if (i == 14)
								magic = VIVA_SUPER_MAGIC;
							else if (i == 30)
								magic = VIVA_SUPER_MAGIC2;
							else
							{
								printf("\n Usage is printed here ---->");
								usage();
							}
							namelen = i;
							dirsize = i+2;
							break;
						default: usage();
					}
				}
			}
		}
	}
	if (!device_name) {
		printf("Device name not specified\n");
		usage();
	}
	psize = partition_size(device_name);
/*	psize = BLOCKS = 131072;*/
	if (BLOCKS < 10) {
		BLOCKS = psize/2;	
		printf("\n  Using partition size = %lu * %d Bytes \n",BLOCKS,VIVA_BLOCK_SIZE);
	}
	else if (BLOCKS > psize) {
		printf("Maximum number of blocks allowable on "); 
		printf("partition %s is %lu\n",device_name,psize);
		usage();
	}
	frag_blocks = (BLOCKS*frag_percent)/100; /* Max no. of fragments allowed */
/* Vamsi: Have to modify this check */
/*	if (BLOCKS > 65536){*/
/*		printf("Too many blocks\n");*/
/*		usage();*/
/*	}*/
	check_mount();		/* is it already mounted? */
	tmp = root_block;
	tmp[0] = 1;
	tmp[1] = 0;
	strcpy(tmp+2,".");
	tmp += dirsize;
	tmp[0] = 1;
	tmp[1] = 0;
	strcpy(tmp+2,"..");
	tmp += dirsize;
	tmp[0] = 2;
	tmp[1] = 0;
	strcpy(tmp+2,".badblocks");
	DEV = open(device_name,O_RDWR );
	if (DEV<0)
		die("unable to open %s");
	if (fstat(DEV,&statbuf)<0)
		die("unable to stat %s");
	if (!S_ISBLK(statbuf.st_mode))
		check=0;
	else if (statbuf.st_rdev == 0x0300 || statbuf.st_rdev == 0x0340)
		die("will not try to make filesystem on '%s'");
	setup_tables();
	fflush(NULL);
	if (check)
		check_blocks();
        else if (listfile)
                get_list_blocks(listfile);
	make_root_inode();
	/*make_bad_inode();*/
	mark_good_blocks();
	write_tables();
	return 0;
}
