/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

/* Functions to keep track of segment counts for inode and zone bitmaps.
If fragments are implemented, segment counts would be modified as in
the functions below. */

static inline void modify_zone_segment_counts(struct super_block *sb,unsigned int seg, unsigned char bitmap_byte, unsigned int modified_bit,int change)  
{
	/*
	Compare byte structure before the new block was allocated
	to the byte structure after allocation, to determine what the
	new segment counts should be.
	Segment counts will always increase if a block is being freed.
	Segment counts will always decrease if a block is being allocated.
	The parameter 'change' should always be +1 or -1. +1 if a block
	is being freed, and -1 if a block is being allocated.
	Number of free one bit boundaries will always change.
	*/

	/* IMPORTANT: In case of block allocation, bitmap_byte should be
	the byte before the bit was set. When blocks are freed however, bitmap_byte
	should be the byte after the bit is set */

	sb->u.viva_sb.s_free_zones[0][seg] += change ;

	/* the only way an 8-bit boundary count can change */
	if (bitmap_byte == 0x00) {
			sb->u.viva_sb.s_free_zones[3][seg] += change; /* 8-bit */
			sb->u.viva_sb.s_free_zones[2][seg] += change; /* 4-bit */
			sb->u.viva_sb.s_free_zones[1][seg] += change; /* 2-bit */

	return;
	}

	/* the only way a 4-bit boundary count can decrease is if
	the modified byte has a free nibble */

	if (((bitmap_byte & 0x0f) == bitmap_byte) && (modified_bit > 3)) {
		/* the leftmost 4 bits (MSB) are free and a bit is 
		being modified among these bits */
		sb->u.viva_sb.s_free_zones[2][seg] += change;
	}
	else if (((bitmap_byte & 0xf0) == bitmap_byte) && (modified_bit < 4)) {
		/* the rightmost 4 bits (LSB) are free and a bit is
		being modified among these bits */	
		sb->u.viva_sb.s_free_zones[2][seg] += change;
	}
	if (((bitmap_byte & 0x3f) == bitmap_byte) && (modified_bit > 5)) {
		/* leftmost 2 bits are free (MSB) */
		sb->u.viva_sb.s_free_zones[1][seg] += change;
	}
	else if (((bitmap_byte & 0xcf) == bitmap_byte) && 
		((modified_bit == 4) || (modified_bit == 5))) {
		/* bits 4 and 5 are free and one of them is being modified */
		sb->u.viva_sb.s_free_zones[1][seg] += change;
	}
	else if (((bitmap_byte & 0xf3) == bitmap_byte) &&
		((modified_bit == 3) || (modified_bit == 2))) {
		/* bits 2 and 3 are free and one of them is being modified */
		sb->u.viva_sb.s_free_zones[1][seg] += change;
	}
	else if (((bitmap_byte & 0xfc) == bitmap_byte) && (modified_bit < 2))	 {
		/* rightmost 2 bits are free (MSB) */
		sb->u.viva_sb.s_free_zones[1][seg] += change;
	}
}

static inline void modify_inode_segment_counts(struct super_block *sb,unsigned int seg, unsigned char bitmap_byte, unsigned int modified_bit,int change)  
{
	/*
	Compare byte structure before the new inode was allocated
	to the byte structure after allocation, to determine what the
	new segment counts should be.
	Segment counts will always increase if a inode is being freed.
	Segment counts will always decrease if a inode is being allocated.
	The parameter 'change' should always be +1 or -1. +1 if a inode
	is being freed, and -1 if a inode is being allocated.
	Number of free 1-bit boundaries will always decrease.
	*/

	/* IMPORTANT: In case of inode allocation, bitmap_byte should be
	the byte before the bit was set. When inodes are freed however, bitmap_byte
	should be the byte after the bit is set */

	sb->u.viva_sb.s_free_inodes[0][seg] += change ;

	/* the only way an 8-bit boundary count can change */
	if (bitmap_byte == 0x00) {
			sb->u.viva_sb.s_free_inodes[3][seg] += change; /* 8-bit */
			sb->u.viva_sb.s_free_inodes[2][seg] += change; /* 4-bit */
			sb->u.viva_sb.s_free_inodes[1][seg] += change; /* 2-bit */
	return;
	}

	/* the only way a 4-bit boundary count can decrease is if
	the modified byte has a free nibble */

	if (((bitmap_byte & 0x0f) == bitmap_byte) && (modified_bit > 3)) {
		/* the leftmost 4 bits (MSB) are free and a bit is 
		being modified among these bits */
		sb->u.viva_sb.s_free_inodes[2][seg] += change;
	}
	else if (((bitmap_byte & 0xf0) == bitmap_byte) && (modified_bit < 4)) {
		/* the rightmost 4 bits (LSB) are free and a bit is
		being modified among these bits */	
		sb->u.viva_sb.s_free_inodes[2][seg] += change;
	}
	if (((bitmap_byte & 0x3f) == bitmap_byte) && (modified_bit > 5)) {
		/* leftmost 2 bits are free (MSB) */
		sb->u.viva_sb.s_free_inodes[1][seg] += change;
	}
	else if (((bitmap_byte & 0xcf) == bitmap_byte) && 
		((modified_bit == 4) || (modified_bit == 5))) {
		/* bits 4 and 5 are free and one of them is being modified */
		sb->u.viva_sb.s_free_inodes[1][seg] += change;
	}
	else if (((bitmap_byte & 0xf3) == bitmap_byte) &&
		((modified_bit == 3) || (modified_bit == 2))) {
		/* bits 2 and 3 are free and one of them is being modified */
		sb->u.viva_sb.s_free_inodes[1][seg] += change;
	}
	else if (((bitmap_byte & 0xfc) == bitmap_byte) && (modified_bit < 2))	 {
		/* rightmost 2 bits are free (MSB) */
		sb->u.viva_sb.s_free_inodes[1][seg] += change;
	}

}

